log_sum_vec <- function(log_vec) {
  max_log_vec <- max(log_vec)
  if (max_log_vec == -Inf) {
    ret <- -Inf
  } else {
    ret <- log(sum(exp(log_vec - max_log_vec))) + max_log_vec
  }
  ret
}
log_rowSums <- function(log_mat) {
  row_maxs <- Rfast::rowMaxs(log_mat, value = TRUE)
  log(Rfast::rowsums(exp(log_mat - row_maxs))) + row_maxs
}
log_colSums <- function(log_mat) {
  col_maxs <- Rfast::colMaxs(log_mat, value = TRUE)
  log(Rfast::rowsums(exp(t(log_mat) - col_maxs))) + col_maxs
}

baum_welch_R <- function(y, 
                         xlabeled, 
                         nstates, 
                         ppi_start, 
                         A_start, 
                         mean_start, 
                         covariances_start, 
                         known_means = NULL,
                         known_covariances = NULL,
                         equal_covariances = FALSE,
                         covariance_structure = "full",
                         y0 = NULL,
                         max_iter = 200, 
                         tol = 1e-4, 
                         silent = TRUE) {
  
  if (!is.null(known_means)) {
    which_known_means <- which(!is.na(known_means))
  }
  if (!is.null(known_covariances)) {
    which_known_covariances <- which(!is.na(known_covariances))
  }
  
  if (equal_covariances) {
    if (length(covariances_start) != 1) {
      stop("If equal_covariances is TRUE, covariances_start must have length 1")
    }
  }
  
  if (!equal_covariances & is.null(known_covariances)) {
      if (length(covariances_start) != nstates) {
        stop("covariances_start must be a list with length equal to nstates")
      }
  }
  
  if (covariance_structure == "diagonal") {
    for (jj in 1:length(covariances_start)) {
      if (sum(abs(covariances_start[[jj]] - diag(diag(covariances_start[[jj]])))) > 0) {
        stop("If covariance_structure is diagonal, covariance_start must contain only diagonal matrices")
      }
    }
  }
  
  states <- sort(unique(xlabeled[!is.na(xlabeled)]))
  if (!identical(states, 1:nstates)) {
    actual_states <- states
    actual_xlabeled <- xlabeled
    states <- seq_along(states)
    for (ss in 1:nstates) {
      xlabeled[xlabeled == actual_states[ss]] <- ss
    }
  } else {
    actual_states <- states
  }
  
  p <- ncol(y)
  TT <- length(xlabeled)

  log_ppi_hat <- log(ppi_start)
  logAhat <- log(A_start)
  means_hat <- mean_start
  if (equal_covariances) {
    covariances_hat <- lapply(1:nstates, function(jj) covariances_start[[1]])
  } else {
    covariances_hat <- covariances_start
  }
  if (!is.null(known_means)) {
    means_hat[which_known_means] <- known_means[which_known_means]
  }
  if (!is.null(known_covariances)) {
    covariances_hat[which_known_covariances] <- known_covariances[which_known_covariances]
  }
  
  log_lik <- -1e30
  log_lik_vec <- log_lik
  loglik_change <- 1e5
  par_change <- 1e5
  iter <- 0
  
  while (iter < max_iter & abs(loglik_change) > tol & par_change > tol) {
    iter <- iter + 1
    logB <- calculate_logB(y = y, means = means_hat, covariances = covariances_hat)
    log_alpha <- log_forward_algorithm_cpp(log_ppi = log_ppi_hat, logA = logAhat, logB = logB, xlabeled = xlabeled)
    log_beta <- log_backward_algorithm_cpp(logA = logAhat, logB = logB, xlabeled = xlabeled)
    log_ab <- log_alpha + log_beta
    
    log_rowsums_log_ab <- log_rowSums(log_ab)
    log_gamma <- log_ab - log_rowsums_log_ab
    gamma <- exp(log_gamma)
    
    log_xi <- log_xi_cpp(log_alpha = log_alpha, logAhat = logAhat,
                         log_beta = log_beta, logB = logB)
    log_xi_mat <- matrix(log_xi, nrow = dim(log_xi)[1])
    
    colSums_gamma <- colSums(gamma)
    
    log_colSums_gamma <- as.numeric(log_colSums_cpp(log_gamma))
    log_colSums_log_xi_mat <- log_colSums_cpp(log_xi_mat)
    logAhat <- matrix(as.numeric(log_colSums_log_xi_mat) - log_colSums_gamma, nrow = nrow(logAhat))
    logAhat <- logAhat - as.numeric(log_rowSums_cpp(logAhat))
    for (ii in 1:nstates) {
      if (sum(is.nan(logAhat[ii, ])) == nstates) {
        logAhat[ii, ] <- rep(- log(nstates), nstates)
      }
    }
    
    if (!(is.null(y0))) {
      log_gamma_calc <- rbind(matrix(c(0, rep(-Inf, nstates - 1)), nrow = nrow(y0), ncol = nstates, byrow = TRUE),
                              log_gamma)
      log_colSums_gamma[1] <- log_sum_vec(c(log_colSums_gamma[1], log(nrow(y0))))
      y_calc <- rbind(y0, y)
    } else {
      log_gamma_calc <- log_gamma
      y_calc <- y
    }
    
    ww <- exp(t(t(log_gamma_calc) - log_colSums_gamma))
    
    means_old <- unlist(means_hat)
    covariances_old <- unlist(covariances_hat)
    
    means_hat <- crossprod(ww, y_calc)
    means_hat <- lapply(1:nstates, function(ii) means_hat[ii, ])
    
    for (ii in 1:nstates) {
      wii <- ww[, ii]
      ycen_ii <- y_calc - t(matrix(means_hat[[ii]], nrow = p, ncol = nrow(y_calc)))
      covariances_hat[[ii]] <- crossprod(ycen_ii * sqrt(wii))  
      if (covariance_structure == "diagonal") {
        covariances_hat[[ii]] <- diag(diag(covariances_hat[[ii]])) 
      }
    }
    if (!is.null(known_means)) {
      means_hat[which_known_means] <- known_means[which_known_means]
    }
    if (equal_covariances) {
      props <- colSums(exp(log_gamma_calc)) / nrow(log_gamma_calc)
      covariance_weighted <- Reduce("+", lapply(1:nstates, function(jj) covariances_hat[[jj]] * props[jj]))
      covariances_hat <- lapply(1:nstates, function(jj) covariance_weighted)
    }
    if (!is.null(known_covariances)) {
      covariances_hat[which_known_covariances] <- known_covariances[which_known_covariances]
    }
    log_ppi_hat <- log_gamma[1, ]
    new_logLik <- log_sum_vec(log_alpha[TT,])
    if (!is.null(y0)) {
      new_logLik <- new_logLik + sum(mvnfast::dmvn(y0, mu = means_hat[[1]], sigma = covariances_hat[[1]], log = TRUE))
    }
    
    means_new <- unlist(means_hat)
    covariances_new <- unlist(covariances_hat)
    
    means_change <- sqrt(sum((means_new - means_old)^2)) / sqrt(sum(means_old^2))
    cov_change <- sqrt(sum((covariances_new - covariances_old)^2)) / sqrt(sum(covariances_old^2))
    par_change <- max(means_change, cov_change)
    
    loglik_change <- new_logLik - log_lik
    log_lik <- new_logLik
    log_lik_vec <- c(log_lik_vec, log_lik)
    if (!silent) {
      print(paste0("log-likelihood value: ", log_lik))
    }
  }
  
  logB <- calculate_logB(y = y, means = means_hat, covariances = covariances_hat)
  log_alpha <- log_forward_algorithm_cpp(log_ppi = log_ppi_hat, logA = logAhat, logB = logB, xlabeled = xlabeled)
  log_beta <- log_backward_algorithm_cpp(logA = logAhat, logB = logB, xlabeled = xlabeled)
  log_ab <- log_alpha + log_beta
  log_rowsums_log_ab <- log_rowSums(log_ab)
  log_gamma <- log_ab - log_rowsums_log_ab
  log_xi <- log_xi_cpp(log_alpha = log_alpha, logAhat = logAhat,
                       log_beta = log_beta, logB = logB)
  log_xi_mat <- matrix(log_xi, nrow = dim(log_xi)[1])
  new_logLik <- log_sum_vec(log_alpha[TT,])
  if (!is.null(y0)) {
    new_logLik <- new_logLik + sum(mvnfast::dmvn(y0, mu = means_hat[[1]], sigma = covariances_hat[[1]], log = TRUE))
  }
  log_lik <- new_logLik
  log_lik_vec <- c(log_lik_vec, log_lik)
  
  
  ll <- new_logLik
  
  npar <-
    nstates - 1 + # pi
    nstates * (nstates - 1) + # A
    nstates * p # mu
  
  if (is.null(known_covariances) & covariance_structure == "full" & !equal_covariances) {
    npar <- npar +
      nstates * p * (p + 1) / 2 # Sigma
  }
  
  if (!is.null(known_covariances) & covariance_structure == "full" & !equal_covariances) {
    how_many_unknown_covariances <- nstates - sum(which_known_covariances)
    npar <- npar +
      how_many_unknown_covariances * p * (p + 1) / 2 # Sigma
  }
  
  if (is.null(known_covariances) & covariance_structure == "diagonal" & !equal_covariances) {
    npar <- npar +
      nstates * p # diagonal Sigma
  }
  
  if (!is.null(known_covariances) & covariance_structure == "diagonal" & !equal_covariances) {
    how_many_unknown_covariances <- nstates - sum(which_known_covariances)
    npar <- npar +
      how_many_unknown_covariances * p # diagonal Sigma
  }
  
  if (is.null(known_covariances) & covariance_structure == "full" & equal_covariances) {
    npar <- npar +
      p * (p + 1) / 2 # equal covariances
  }
  
  if (!is.null(known_covariances) & covariance_structure == "full" & equal_covariances) {
    how_many_unknown_covariances <- nstates - sum(which_known_covariances)
    npar <- npar +
      0 # it should return an error if you provide more than one covariance if equal_covariances is TRUE
  }
  
  if (is.null(known_covariances) & covariance_structure == "diagonal" & equal_covariances) {
    npar <- npar +
      p # equal diagonal covariances
  }
  
  if (!is.null(known_covariances) & covariance_structure == "diagonal" & equal_covariances) {
    how_many_unknown_covariances <- nstates - sum(which_known_covariances)
    npar <- npar +
      0 # it should return an error if you provide more than one covariance if equal_covariances is TRUE
  }
  
  AIC <- -2 * ll + 2*npar
  BIC <- -2 * ll + log(length(xlabeled))*npar
  if (!is.null(y0)) {
    BIC <- -2 * ll + log(length(xlabeled) + nrow(y0))*npar
  }
  
  
  if (!identical(actual_states, 1:nstates)) {
    xlabeled <- actual_xlabeled
  }
  
  out <- list(y = y,
              xlabeled = xlabeled,
              log_lik = log_lik,
              log_lik_vec = log_lik_vec,
              iter = iter,
              logB = logB,
              log_alpha = log_alpha,
              log_beta = log_beta,
              log_gamma = log_gamma,
              log_xi = log_xi,
              logAhat = logAhat,
              means_hat = means_hat,
              covariances_hat = covariances_hat,
              log_pi_hat = log_ppi_hat,
              AIC = AIC,
              BIC = BIC)
  
}

# entropy density of sequence ---------------------------------------------
entropy_of_sequence_hmm <- function(hmm, xlabeled, y, window) {
  
  log_ppi <- get_logp_start(hmm, xlabeled, 1)
  logAhat <- hmm$logAhat
  nstates <- ncol(hmm$logB)
  logB <- calculate_logB(
    y = y, 
    means = hmm$means_hat, 
    covariances = hmm$covariances_hat)
  
  if (all(window == nrow(y))) {
    log_alpha <- log_forward_algorithm_cpp(log_ppi, logAhat, logB, xlabeled)[window, , drop = FALSE]
    log_gamma <- log_alpha - log_rowSums(log_alpha)
  } else {
    log_alpha <- log_forward_algorithm_cpp(log_ppi, logAhat, logB, xlabeled)[window, , drop = FALSE]
    log_beta <- log_backward_algorithm_cpp(logAhat, logB, xlabeled)[window, , drop = FALSE]
    log_ab <- log_alpha + log_beta
    log_rowsums_log_ab <- log_sum_vec(log_ab)
    log_gamma <- log_ab - log_rowsums_log_ab
  }
  
  ntest <- length(window)
  
  log_cc <- matrix(0, nrow = ntest, ncol = nstates)
  log_cc1 <- log_gamma[1, ]
  
  
  log_cc[1, ] <- log_cc1 - log_sum_vec(log_cc1)
  if (ntest > 1) {
    for (tt in 2:ntest) {
      if (!is.na(xlabeled[window[tt]])) {
        log_cc[tt, xlabeled[window[tt]]] <- 0
        log_cc[tt, -xlabeled[window[tt]]] <- -Inf
      } else {
        log_cc_tt <- log_colSums_cpp(log_cc[tt-1, ] + t(t(logAhat) + logB[window[tt], ]))
        log_cc[tt, ] <- log_cc_tt - log_sum_cpp(log_cc_tt)
      }
    }
  }
  
  # log_p_back
  if (ntest > 1) {
    log_p_back <- array(0, dim = c(ntest - 1, nstates, nstates))
    if (ntest > 2) {
      for (tt in 2:ntest) {
        log_p_back_ttm1 <- logAhat + log_cc[tt - 1, ]
        log_p_back[tt-1, , ] <- t(t(log_p_back_ttm1) - as.numeric(log_colSums(log_p_back_ttm1)))
      }
    }
  }
  
  H <- matrix(NA, nrow = ntest, ncol = nstates)
  H[1, ] <- 0
  if (ntest > 1) {
    for (tt in 2:ntest) {
      possible_states <- 1:nstates
      if (!is.na(xlabeled[window[tt]])) {
        H[tt, -xlabeled[window[tt]]] <- 0
        possible_states <- xlabeled[window[tt]]
      }
      for (jj in possible_states) {
        term1 <- sum(H[tt - 1, ] * exp(log_p_back[tt - 1, , jj]))
        if (max(log_p_back[tt - 1, , jj]) == 0) {
          term2 <- 0
        } else {
          term2 <- sum(exp(log_p_back[tt - 1, , jj]) * (log_p_back[tt - 1, , jj]))
        }
        H[tt, jj] <- term1 - term2
      }
    }
  }
  
  term1 <- H[ntest, ] * exp(log_cc[ntest, ])
  if (max(log_cc[ntest, ]) == 0) {
    term2 <- 0
  } else {
    term2 <- exp(log_cc[ntest, ]) * log_cc[ntest, ]
  }
  entropy_of_sequence <- sum(term1 - term2)
  return(entropy_of_sequence)
}

log_density_of_sequence <- function(hmm, xlabeled, y, window) {
  
  log_ppi <- get_logp_start(hmm, xlabeled, window)
  
  length_x <- length(xlabeled)
  length_y <- nrow(y)
  
  logB <- calculate_logB(y = y[window - (length_x - length_y), , drop = FALSE], means = hmm$means_hat, covariances = hmm$covariances_hat)
  log_alpha <- log_forward_algorithm_cpp(log_ppi = log_ppi, logA = hmm$logAhat, logB = logB, xlabeled = xlabeled[window])
  ret <- log_sum_cpp(last(log_alpha))
  return(ret)
}


get_logp_start <- function(hmm, xlabeled, window) {
  
  tstart <- min(window)
  nstates <- ncol(hmm$logB)
  
  logAhat <- hmm$logAhat
  log_p_x <- matrix(NA, nrow = tstart, ncol = nstates)
  if (!is.na(xlabeled[1])) {
    log_p_x[1, ] <- -Inf
    log_p_x[1, xlabeled[1]] <- 0
  } else {
    log_p_x[1, ] <- hmm$log_pi_hat
  }
  if (tstart >= 2) {
    for (tt in 2:tstart) {
      if (!is.na(xlabeled[tt])) {
        log_p_x[tt, ] <- -Inf
        log_p_x[tt, xlabeled[tt]] <- 0
      } else {
        log_p_x[tt, ] <- log_colSums_cpp(logAhat + log_p_x[tt - 1, ])
      }
    }
  }
  
  return(log_p_x[nrow(log_p_x), ])
  
}


calculate_logB <- function(y, means, covariances) {
  nstates <- length(means)
  logBlist <- lapply(1:nstates, function(ii) {
      log_density <- mvnfast::dmvn(y, mu = means[[ii]], sigma = covariances[[ii]], log = TRUE)
    return(log_density)
  })
  logB <- do.call(cbind, logBlist)
  return(logB)
}

# simulate from hmm model
simulate_hmm <- function(n, log_ppi, logA, means, covariances, xlabeled, n_last_y = n) {
  
  if (n != length(xlabeled)) {
    stop("length of xlabeled must be equal to n")
  }
  
  which_label <- which(!is.na(xlabeled))
  nstates <- length(log_ppi)
  TT <- length(xlabeled)
  p <- length(means[[1]])
  sequence_list <- list()
  cut_points <- unique(c(1, which_label, TT))
  log_gamma <- matrix(NA, nrow = TT, ncol = nstates)
  log_alpha_list <- list()
  log_beta_list <- list()
  
  xsim <- rep(NA, TT)
  xsim[which_label] <- xlabeled[which_label]
  
  if (n == 1) {
    if (!is.na(xlabeled)) {
      xsim <- xlabeled
    } else {
      xsim <- sample(1:nstates, 1, prob = exp(log_ppi))
      ysim <- mvnfast::rmvn(1, mu = means[[xsim[ii]]], sigma = covariances[[xsim[[ii]]]])
      ret <- list(x = xsim, y = ysim)
      return(ret)
    }
  }
  
  A <- exp(logA)
  
  for (ii_known in 1:(length(cut_points) - 1)) {
    sequence_list[[ii_known]] <- cut_points[ii_known]:cut_points[ii_known + 1]
    start_ii <- sequence_list[[ii_known]][1]
    end_ii <- max(sequence_list[[ii_known]])
    TT_ii <- length(sequence_list[[ii_known]])
    
    if (is.na(xlabeled[end_ii])) { # Cases 1 and 2 (final x unknown)
      if (is.na(xlabeled[start_ii])) { # Case 1 (also first x unknown)
        ppi <- exp(log_ppi)
        xsim <- simulate_state_sequence(ppi, TT_ii, A)
      } else {
        ppi <- rep(0, length(log_ppi))
        ppi[xlabeled[start_ii]] <- 1
      }
      
      xsim[start_ii:end_ii] <- simulate_state_sequence(ppi, TT_ii, A)
    } else { # Cases 3 and 4 (final x known)
      
      if (is.na(xlabeled[start_ii])) { # Case 4 (first x unknown)
        log_alpha_ii <- matrix(NA, nrow = TT_ii, ncol = nstates)
        log_beta_ii <- matrix(NA, nrow = TT_ii, ncol = nstates)
        log_gamma_ii <- matrix(NA, nrow = TT_ii, ncol = nstates)
        
        log_alpha_ii[1, ] <- log_ppi
        for (tt in 2:TT_ii) {
          log_alpha_ii[tt, ] <- log_colSums_cpp(log_alpha_ii[tt - 1, ] + logA)
        }
        log_beta_ii[TT_ii - 1, ] <- logA[, xlabeled[end_ii]]
        if (TT_ii > 2) {
          for (tt in (TT_ii - 2):1) {
            log_beta_ii[tt, ] <- log_colSums_cpp(t(logA) + log_beta_ii[tt + 1, ])
          }
        }
        log_gamma_ii <- log_alpha_ii + log_beta_ii
        log_gamma_ii <- log_gamma_ii - as.numeric(log_rowSums_cpp(log_gamma_ii))
        
        xsim[start_ii] <- sample(1:nstates, 1, prob = exp(log_gamma_ii[1, ]))
      }
      
      # if TT_ii == 2 there are no unknown states
      if (TT_ii > 2) {
        # Case 3 (first and last x known)
        for (ttii in 1:(TT_ii - 2)) {
          # print(ttii)
          
          current_tt <- sequence_list[[ii_known]][ttii]
          current_length <- length(ttii:TT_ii)
          log_alpha_ii <- matrix(NA, nrow = current_length, ncol = nstates)
          log_beta_ii <- matrix(NA, nrow = current_length, ncol = nstates)
          log_gamma_ii <- matrix(NA, nrow = current_length, ncol = nstates)
          
          log_alpha_ii[1, ] <- -Inf
          log_alpha_ii[1, xsim[current_tt]] <- 0
          for (ttkk in 2:nrow(log_alpha_ii)) {
            log_alpha_ii[ttkk, ] <- log_colSums_cpp(log_alpha_ii[ttkk - 1, ] + logA)
          }
          log_beta_ii[current_length - 1, ] <- logA[, xlabeled[end_ii]]
          for (tt in (current_length - 2):1) {
            log_beta_ii[tt, ] <- log_colSums_cpp(t(logA) + log_beta_ii[tt + 1, ])
          }
          log_gamma_ii <- log_alpha_ii + log_beta_ii
          log_gamma_ii <- log_gamma_ii - as.numeric(log_rowSums_cpp(log_gamma_ii))
          
          xsim[current_tt + 1] <- sample(1:nstates, 1, prob = exp(log_gamma_ii[2, ]))
          
        }
      }
    }
  }
  idx <- (TT - n_last_y + 1):TT
  ysim <- lapply(idx, function(ii) mvnfast::rmvn(1, mu = means[[xsim[ii]]], sigma = covariances[[xsim[[ii]]]]))
  ysim <- do.call(rbind, ysim)
  
  ret <- list(x = xsim, y = ysim)
  return(ret)
}

simulate_hmm_with_constant_length_sequences <- function(TT = 1000, 
                                                        T_out = 5, 
                                                        T_min_IC = 60,
                                                        T_max_IC = 85,
                                                        log_ppi1 = c(0, -Inf), 
                                                        log_ppi2 = c(0, -Inf, -Inf), 
                                                        logA1 = log(rbind(c(0.97, 0.03), c(0.2, 0.8))), 
                                                        logA2 = log(rbind(c(0.97, 0.015, 0.015), c(0.2, 0.8, 0), c(0.2, 0, 0.8))), 
                                                        means = list(rep(0, p), c(1.5, rep(0, p - 1)), c(-1.5, rep(0, p - 1))), 
                                                        covariances = list(diag(p), diag(p), diag(p))) {
  
  N1 <- length(log_ppi1)
  N2 <- length(log_ppi2)
  TT1 <- as.integer(TT/2)
  TT2 <- TT - TT1
  
  x <- rep(NA, TT)
  tt <- 1
  while(tt <= TT1) {
    tt_start <- tt
    tt_end <- tt_start + sample(T_min_IC:T_max_IC, 1)
    tt_end <- min(tt_end, TT1)
    x[tt_start:tt_end] <- 1
    tt_start <- tt_end + 1
    tt_end <- tt_start + T_out - 1
    if (tt_end < TT1) {
      x[tt_start:tt_end] <- 2
    } else {
      x[tt_start:TT1] <- 1
    }
    
    tt <- tt_end + 1
  }
  tt <- TT1 + 1
  while(tt <= TT) {
    tt_start <- tt
    tt_end <- tt_start + sample(T_min_IC:T_max_IC, 1)
    tt_end <- min(tt_end, TT)
    x[tt_start:tt_end] <- 1
    tt_start <- tt_end + 1
    tt_end <- tt_start + T_out - 1
    if (tt_end < TT) {
      x[tt_start:tt_end] <- 3
    } else {
      x[tt_start:TT] <- 1
    }
    tt <- tt_end + 1
  }
  x <- x[1:TT]
  
  y <- lapply(1:TT, function(ii) mvnfast::rmvn(1, mu = means[[x[ii]]], sigma = covariances[[x[[ii]]]]))
  y <- do.call(rbind, y)
  
  ret <- list(x = x, y = y)
  return(ret)
}


get_classification_scores <- function(xhat, x) {
  
  xhat[xhat > 1] <- 2
  x[x > 1] <- 2
  
  nclasses <- length(table(c(x, xhat)))
  
  accuracy <- mean(xhat == x) # (TP+TN) / ALL
  
  precision <- recall <- f1 <- auc <- numeric(nclasses)
  
  if (nclasses > 2) {
    for (jj in 1:nclasses) {
      precision[jj] <- mean(x[xhat == jj] == jj) # TP / PREDICTED_P
      recall[jj] <- mean(xhat[x == jj] == jj) #TP / ALL_P
      f1[jj] <- 1 / mean(1 / precision[jj], 1 / recall[jj])
      xjj <- ifelse(x == jj, 1, 2)
      xjjhat <- ifelse(xhat == jj, 1, 2)
      auc[jj] <- as.numeric(auc(roc(xjj, xjjhat, quiet = TRUE)))
    }
  }
  
  if (nclasses == 2) { # assuming 2 is minority, positive class
    precision <- mean(x[xhat == 2] == 2) # TP / PREDICTED_P
    recall <- mean(xhat[x == 2] == 2) #TP / ALL_P
    f1 <- 1 / mean(c(1 / precision, 1 / recall))
    auc <- as.numeric(auc(roc(x, as.numeric(xhat), quiet = TRUE)))
  }
  
  list(accuracy = accuracy,
       precision = precision,
       recall = recall,
       f1 = f1,
       auc = auc)
  
}

are_equal <- function(vec1, vec2) {
  identical(as.numeric(vec1), as.numeric(vec2))
}


baum_welch_R_with_initialization <- function(y = y,
                                             xlabeled = xlabeled,
                                             equal_covariances = TRUE,
                                             tol = 1e-3,
                                             use_true = FALSE,
                                             max_nstates = 5,
                                             k = 5,
                                             ntry = 10,
                                             select_nstates_manually = NULL,
                                             true_mean,
                                             true_covariance,
                                             y0 = NULL) {
  
  
  nstates_min <- length(table(xlabeled))
  if (!(is.null(y0))) {
    y_all <- rbind(y0, y)
  } else {
    y_all <- y
  }
  # Try using CovRobust, if it fails, use CovMcd
  tryCatch({
    cov_robust <- rrcov::CovRobust(y_all)
  }, error = function(e) {
    cov_robust <- rrcov::CovMcd(y_all, alpha = 0.9)
  })
  mean_start <- list(cov_robust$center)
  covariances_start <- list(cov_robust$cov)
  nstates <- 1
  
  if (nstates_min <= 1) {
    # 1 state
    mod1 <- baum_welch_R(y = y,
                         xlabeled = xlabeled,
                         nstates = 1,
                         ppi_start = 1,
                         A_start = matrix(1),
                         mean_start = list(cov_robust$center),
                         covariances_start = list(cov_robust$cov),
                         equal_covariances = TRUE,
                         tol = tol,
                         y0 = y0
    )
  } else {
    mod1 <- list(AIC = 1e10)
  }
  
  if (max_nstates >= 2) {
    # 2 states
    ycen <- t(y_all) - cov_robust$center
    ycen_mean <- if (k == 1) ycen else t(caTools::runmean(t(ycen), k = k, align = "right", endrule = "NA"))
    T2 <- colSums(ycen_mean * (solve(cov_robust$cov) %*% ycen_mean))
    order_T2 <- order(-T2)
    if (!is.null(y0)) {
      order_T2 <- order_T2[order_T2 > nrow(y0)]
      ntry <- min(sum(!is.na(T2[-(1:nrow(y0))])), ntry)
    } else {
      ntry <- min(sum(!is.na(T2)), ntry)
    }
    
    
    
    best_mod2 <- list(log_lik = -1e5)
    log_lik_try <- numeric(ntry)
    niter_try <- numeric(ntry)
    A_start <- cbind(c(0.99, 0.01), c(0.01, 0.99))
    for (ii in 1:ntry) {
      mean_start[[2]] <- ycen_mean[, order_T2[ii]]
      mod2 <- baum_welch_R(y = y,
                           xlabeled = ifelse(xlabeled <= 2, xlabeled, NA),
                           nstates = 2,
                           ppi_start = c(1, 0),
                           A_start = A_start,
                           mean_start = mean_start,
                           covariances_start = covariances_start,
                           equal_covariances = TRUE,
                           tol = tol,
                           y0 = y0)
      log_lik_try[ii] <- mod2$log_lik
      niter_try[ii] <- mod2$iter
      if (mod2$log_lik > best_mod2$log_lik) best_mod2 <- mod2
    }  
    mod2 <- best_mod2
    mod_true <- baum_welch_R(y = y,
                             xlabeled = ifelse(xlabeled <= 2, xlabeled, NA),
                             nstates = 2,
                             ppi_start = c(1, 0),
                             A_start = A_start,
                             mean_start = true_mean[1:2],
                             covariances_start = true_covariance,
                             equal_covariances = TRUE,
                             tol = tol,
                             y0 = y0)
    if (use_true & mod_true$log_lik > mod2$log_lik) {
      mod2 <- mod_true
    } 
    
    nstates <- which.min(c(mod1$AIC, mod2$AIC))
    mod <- mod1
  } else {
    mod2 <- list(AIC = 1e10)
  }
  
  
  if ((nstates_min <= 3 & nstates == 2 & max_nstates >= 3) | nstates_min >= 3) {
    mod <- mod2
    # 3 states
    ycen2 <- t(y_all) - mod2$means_hat[[2]]
    ycen_mean2 <- if (k == 1) ycen2 else t(caTools::runmean(t(ycen2), k = k, align = "right", endrule = "NA"))
    
    # ycen_mean <- ycen
    # T2 <- colSums(ycen_mean * (solve(cov_robust$cov) %*% ycen_mean))
    T2_2 <- colSums(ycen_mean2 * (solve(mod2$covariances_hat[[1]]) %*% ycen_mean2))
    T2_12 <- T2
    ycen_mean_12 <- ycen_mean
    which_near_to_2 <- which(T2_2 < T2)
    if (length(which_near_to_2) > 0) {
      T2_12[which_near_to_2] <- T2_2[which_near_to_2]
      ycen_mean_12[, which_near_to_2] <- ycen_mean2[, which_near_to_2]
    }
    
    best_mod3 <- list(log_lik = -1e5)
    log_lik_try <- numeric(ntry)
    niter_try <- numeric(ntry)
    A_start <- cbind(c(0.98, 0.01, 0.01), c(0.01, 0.98, 0.01), c(0.01, 0.01, 0.98))
    mean_start <- mod2$means_hat
    mean_start[[1]] <- cov_robust$center
    
    order_T2_12 <- order(-T2_12)
    if (!is.null(y0)) {
      order_T2_12 <- order_T2_12[order_T2_12 > nrow(y0)]
    }

    for (ii in 1:ntry) {
      mean_start[[3]] <- ycen_mean_12[, order_T2_12[ii]]
      mod3 <- baum_welch_R(y = y,
                           xlabeled = xlabeled,
                           nstates = 3,
                           ppi_start = c(1, 0, 0),
                           A_start = A_start,
                           mean_start = mean_start,
                           covariances_start = covariances_start,
                           equal_covariances = TRUE,
                           tol = tol,
                           y0 = y0)
      log_lik_try[ii] <- mod3$log_lik
      niter_try[ii] <- mod3$iter
      if (mod3$log_lik > best_mod3$log_lik) best_mod3 <- mod3
    }  
    mod3 <- best_mod3
    mod_true <- baum_welch_R(y = y,
                             xlabeled = xlabeled,
                             nstates = 3,
                             ppi_start = c(1, 0, 0),
                             A_start = A_start,
                             mean_start = true_mean[1:3],
                             covariances_start = true_covariance,
                             equal_covariances = TRUE,
                             tol = tol,
                             y0 = y0)
    if (use_true & mod_true$log_lik > mod3$log_lik) {
      mod3 <- mod_true
    } 
    nstates <- which.min(c(mod1$AIC, mod2$AIC, mod3$AIC))
  } else {
    mod3 <- list(AIC = 1e10)
  }
  
  
  nstates <- which.min(c(mod1$AIC, mod2$AIC, mod3$AIC))
  nstates <- max(nstates, nstates_min)
  if (!is.null(select_nstates_manually)) {
    nstates <- select_nstates_manually
  }
  if (nstates == 1) mod <- mod1
  if (nstates == 2) mod <- mod2
  if (nstates == 3) mod <- mod3
  
  return(mod)
  
}